from ..util import *

def io_readline(line, read_token):
    sline = line.strip()

    is_readline = False
    if config.READ_KEYWORD and read_token.startswith('read'):
        is_readline = True
    if read_token.startswith('raw_input'):
        is_readline = True
    if is_readline:
        args = smart_split(sline[len(read_token):], ' ,')
    else:
        args = smart_split(sline[len(read_token):], ' ,')

    # if the line already has >> or << on it, we don't process it
    for arg in args:
        if arg == "<<" or arg == ">>":
            return line

    tokens = []
    cin_tokens = []
    cout_tokens = []
    for arg in args:
        if arg[0] != '"':
            if cout_tokens:
                tokens.append("cout")
                tokens.extend(cout_tokens)
                tokens.append(';')
                cout_tokens = []

            cin_tokens.append(">>")
            cin_tokens.append(arg)
        else:
            if cin_tokens:
                tokens.append("cin")
                tokens.extend(cin_tokens)
                tokens.append(';')
                cin_tokens = []

            cout_tokens.append("<<")
            cout_tokens.append(arg)

    if cout_tokens:
        tokens.append("cout")
        tokens.extend(cout_tokens)
        tokens.append(';')

    if cin_tokens:
        tokens.append("cin")
        tokens.extend(cin_tokens)

    line = " ".join(tokens)
    return line

def io_printline(line, indent):
    sline = line.strip()
    print_token = None
    # PRINT NO NEWLINE
    add_space = True
    if sline.startswith('!!'):
        print_token = "!! "
    if sline.startswith('puts '):
        print_token = 'puts '
        if sline.endswith(","):
            add_space = False

    if print_token:
        args = smart_split(sline[len(print_token):], ' ,')
        line = "%scout << %s" % (' ' * indent, " << ".join(args))
        if add_space:
            line = '%s << " "' % (line)
        return line

    # PRINT WITH NEWLINE
    if sline == "print":
        return "%scout << endl" % (' ' * indent)

    for tok in ["!", "std::cout ", "cout ", "print "]:
        if sline.startswith(tok):
            if tok == "print ":
                args = smart_split(sline[len(tok):], ',')
            else:
                args = smart_split(sline[len(tok):], ' ,')
            no_add = False
            for arg in args:
                if arg == "<<":
                    no_add = True
                    break

            if no_add:
                continue

            spc = "endl";
            if sline.endswith(','):
                spc = "' '";

            if not args:
                line = "%scout << %s" % (' ' * indent, spc)
            else:
                line = "%scout << %s << %s" % (' ' * indent, " << ' ' << ".join(args), spc)

    return line

def replace_io_keywords(lines):
    new_lines = []
    tokens = [ '?', 'read', '??', 'cin', 'raw_input']
    for line in lines:
        if ignore_line(line, new_lines):
            continue
        indent = get_indent(line)
        sline = line.strip()

        read_token = None
        toks = smart_split(line, '(): ')

        for i, t in enumerate(toks):
            for tok in tokens:
                if i > 0 and (tok[0] == '?'):
                    continue

                if t == tok:
                    read_token = tok
                    rtoks = toks[i:]
                    rline = " ".join(rtoks)
                    pline = toks[:i]

        if read_token:
            if len(pline):
                pline = " ".join(pline) + " "
            else:
                pline = ""

            rline = io_readline(rline, read_token)
            line = "%s%s%s" % (' ' * indent, pline, rline)

        else:
            line = io_printline(line, indent)


        new_lines.append(line)

    return new_lines



import re

# from stack overflow somewhere
def comment_remover(text):
    def replacer(match):
        s = match.group(0)
        if s.startswith('/'):
            return " " # note: a space and not an empty string
        else:
            return s

    pattern = re.compile(
        r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"',
        re.DOTALL | re.MULTILINE
    )
    return re.sub(pattern, replacer, text)


def skip_comments(lines):
    return comment_remover(''.join(lines)).split('\n')




from ..util import *


def add_trailing_semicolons(lines):
    new_lines = []
    indents = []
    for line in lines:
        indents.append(get_indent(line))

    indents.append(0)
    double_ignore = IGNORE_CHAR + IGNORE_CHAR

    for i, line in enumerate(lines):
        line = line.rstrip();
        cline = line.strip()

        # we dont ignore lines with preceding semicolons
        if ignore_line(line, new_lines):
            continue
        if cline == double_ignore:
            new_lines.append(line)
            continue

        if not cline or cline[0] == '#':
            new_lines.append(line)
            continue

        j = i+1
        while j < len(lines) and not lines[j].strip():
            j += 1

        if indents[i] < indents[j]:
            new_lines.append(line)
            continue

        add_semi = True
        # there are a lot of reasons not to add a semicolon, like
        # if the line doesn't end with a backslash, colon or start with class keyword
        # or if the line has a template< decl on it
        if cline[-1] == '\\':
            add_semi = False
        elif cline[-1] == ':':
            add_semi = False
        elif cline[-1] == ';':
            add_semi = False
        elif cline[-1] == '{':
            add_semi = False
        elif cline.startswith('class '):
            add_semi = False
        elif line_is_template(cline):
            add_semi = False

        if add_semi:
            # we have `do` here, just in case
            for tok in ['for ', 'while ', 'do ', 'else if', 'if ']:
                if cline.startswith(tok):
                    add_semi = False
                    break

        if add_semi:
            line += ';'

        new_lines.append(line)

    return new_lines

def add_curly_braces(lines):
    new_lines = []
    indent_levels = [0]
    nb = 0

    outer_indent = 0

    for i, line in enumerate(lines):
        line = line.rstrip('\n')
        indent = get_indent(line)
        if not line.strip():
            new_lines.append(line)
            continue

        if ignore_line(line, new_lines):
            continue

        if visibility_line(line):
            # TODO: when we hit a public line, i guess we reset indent levels?
            new_lines.append(line)
            continue

        if hash_line(line):
            new_lines.append(line)
            continue

        if is_class(line):
            outer_indent = indent

        if indent_levels[-1] > indent:
            while indent_levels[-1] > indent:
                indent_levels.pop()
                new_lines[nb] += ' }'

                if indent_levels[-1] == outer_indent and new_lines[nb][-1] != ';':
                    new_lines[nb] += ';'


        if indent_levels[-1] < indent:
            if not visibility_line(new_lines[nb]):
                indent_levels.append(indent)
                if not case_statement(new_lines[nb]):
                    new_lines[nb] = new_lines[nb].rstrip(':')
                new_lines[nb] += ' {'

        new_lines.append(line)

        # last non blank line is this one
        nb = i

    while indent_levels:
        if indent_levels[-1] > 0:
            new_lines[nb] += ' }'

        if len(new_lines[nb]) and new_lines[nb][0] != "#" and new_lines[nb][-1] != ";":
            new_lines[nb] += ';';

        indent_levels.pop()
    new_lines.append('')

    return new_lines

def join_backslash_lines(lines):
    new_lines = ['']

    i = 0
    full_line = []
    while i < len(lines):
        line = lines[i].rstrip()
        if full_line:
            line = line.strip()
        if line.endswith('\\'):
            # strip off the backslash, strip off the remaining spaces
            full_line.append(line.rstrip('\\').rstrip())
        else:
            if full_line:
                full_line.append(line.strip())
            else:
                full_line.append(line)
            new_lines.append(' '.join(full_line))
            full_line = []

        i += 1

    return new_lines



def add_preceding_ignore_chars(lines):
    new_lines = []
    in_ignore_block = False
    ig = IGNORE_CHAR
    for line in lines:
        if line.strip().startswith('```'):
            in_ignore_block = not in_ignore_block
        elif in_ignore_block:
            new_lines.append('%s%s%s' % (ig, ig, line))
        else:
            new_lines.append(line)

    return new_lines

def remove_preceding_ignore_chars(lines):
    new_lines = []
    double_ignore = IGNORE_CHAR + IGNORE_CHAR
    for line in lines:
        indent = get_indent(line)
        sline = line.strip()

        if sline.find(double_ignore) != -1:
            line = line.replace(double_ignore, '', 1)

        if sline and sline[0] == IGNORE_CHAR:
            line = line.replace(IGNORE_CHAR, ' ', 1)

        new_lines.append(line)

    return new_lines


# we might have a case like: if (a) || (b): which still needs parentheses
# the only good case is when the first paren is matching the last paren
def needs_parens(nline):
    if nline[0] != '(' or nline[-1] != ')':
        return True

    paren_stack = []
    last_pop = 0
    for i, c in enumerate(nline):
        if c == '(':
            paren_stack.append(i)
        elif c == ')':
            last_pop = paren_stack.pop()

    if paren_stack or last_pop != 0:
        return True


def add_parentheses(lines):
    new_lines = []
    replace = ["if ", "while ", "for ", "else if ", "switch " ]
    for line in lines:
        if ignore_line(line, new_lines):
            continue

        indent = get_indent(line)
        sline = line.strip()
        # TODO: join multiple lines (maybe look for next colon?)
        for tok in replace:
            if sline.startswith(tok):
                nline = line[indent+len(tok):].rstrip(':')
                if nline[-1] != ';' and needs_parens(nline):
                    line = "%s%s(%s) " % (' ' * indent, tok, nline)

        new_lines.append(line)

    return new_lines




from __future__ import print_function
from .util import *

import os

from . import id_recognizer
def read_scopings(lines):
    indent_levels = [0]
    nb = 0
    scopings = {}
    scope = {}
    scope_stack = { 0: scope }


    for i, line in enumerate(lines):
        line = line.rstrip('\n')
        indent = get_indent(line)

        if indent in scope_stack:
            scope = scope_stack[indent]
        else:
            scope = {}
            scope_stack[indent] = scope

        scopings[i] = [c for c in scope]

        if not line:
            continue

        if visibility_line(line):
            continue

        # last non blank line is now this one
        nb = i

        if indent_levels[-1] > indent:
            while indent_levels[-1] > indent:
                indent_levels.pop()
                scope = scope_stack[indent_levels[-1]]


        if indent_levels[-1] < indent:
            if not visibility_line(lines[nb]):
                indent_levels.append(indent)

                scope = dict([(c, c) for c in scope])
                scope_stack[indent] = scope

        if i < len(lines) - 1:
            x = 1
            # skip empty-ish lines when figuring out next indent
            while i+x < len(lines) -1 and not lines[i+x].strip() :
                x += 1

            next_line = lines[i+x]
            next_indent = get_indent(next_line)

            if indent < next_indent:
                scope = dict([(c, c) for c in scope])
                scope_stack[next_indent] = scope

        new = id_recognizer.add_identifiers(line, scope)

        # compare libclang vs. our own version
        # import clanger
        # new2 = clanger.add_identifiers(line, bscope)
        # debug("TOKENS", line, new, new2)

    return scopings

# after line is already decorated, we can check if its a function
# so we can make a .h file for this .cpp file
CONTROL_WORDS = [
    'else',
    'if',
    'class',
    'struct',
    'typedef',
    'for',
    'while']

def line_is_function(line):
    args = smart_split(line, ' ')
    if not args:
        return

    if args[0] in CONTROL_WORDS:
        return False

    if args[-1].endswith('{') and line.find('(') != -1:
        return True

    return False

def line_is_class_definition(line):
    args = smart_split(line, ' ')
    if not args:
        return

    if args[0].find('::') != -1 or args[1].find('::') != -1:
        return True

def file_contains_main(lines):
    i = 0
    while i < len(lines):
        line = lines[i]
        if is_struct(line) or is_class(line):
            until = extract_until_close(lines, i)
            if until:
                i = until
                continue


        if re.search("main(.*) {", line):
            return True

        i += 1

    return False

def gather_includes(file, includes, base_dir=None):
    file_path = os.path.join(base_dir or '', file)
    if file_path in includes:
        return

    dirname = os.path.dirname(file)
    fname = os.path.basename(file)

    if base_dir == None:
        base_dir = dirname
    else:
        base_dir = os.path.join(base_dir, dirname)

    if not os.path.exists(file_path):
        print("missing file:", file)
        return

    file = os.path.join(base_dir, fname)
    includes[file_path] = True


    with open(file_path) as f:
        lines = f.readlines()

    for line in lines:
        cline = line.strip()
        if cline.startswith("#include"):
            tokens = cline.split()
            include = tokens[1]
            if include[0] != '"':
                continue
            include = include.strip('"')

            fname, ext = os.path.splitext(include)
            cpyname = "%s.cpy" % fname
            if os.path.exists(os.path.join(base_dir, cpyname)):
                gather_includes(cpyname, includes, base_dir)
            else:
                gather_includes(include, includes, base_dir)

    return includes

def gather_files(files):
    ret = {}
    original_dir = os.getcwd()
    for file in files:
        if file == '-':
            ret[file] = True
            continue

        if not os.path.exists(file):
            continue
        file_dir, file = os.path.split(file)
        gather_includes(file, ret,file_dir)

    os.chdir(original_dir)
    return list(ret.keys())

def guess_required_files(lines):
    requires = set()
    defines = set()

    REQUIRE_KEYWORDS = {
        "<iostream>" : [ "cout", "cin", "endl" ],
        "<vector>" : [ "vector" ],
        "<tuple>" : [ "tuple", "make_tuple", "tie", "std::tie"],
        "<queue>" : [ "queue" ],
        "<deque>" : [ "deque" ],
        "<map>" : ["map"],
        "<unordered_map>" : ["unordered_map"],
        "<cstdio>" : ["printf", "scanf"],
        "<memory>" : ["shared_ptr", "unique_ptr", "weak_ptr"]
    }

    DEFINE_KEYWORDS = {
        "len(x) (int)(x).size()" : [ "len" ]
    }


    def add_requires(line_tokens, require_tokens, require):
        if not require in requires:
            for tok in require_tokens:
                if tok in line_tokens:
                    requires.add(require)


    def add_defines(line_tokens, define_tokens, word):
        if not word in defines:
            for tok in define_tokens:
                if tok in line_tokens:
                    defines.add(word)


    using_namespace_std = False
    for line in lines:
        if not using_namespace_std and line.find("using namespace std;") != -1:
            using_namespace_std = True

        tokens = smart_split(line, " :<>()")
        for require in REQUIRE_KEYWORDS:
            add_requires(tokens, REQUIRE_KEYWORDS[require], require)

        for define in DEFINE_KEYWORDS:
            add_defines(tokens, DEFINE_KEYWORDS[define], define)


    prepend = []
    for r in requires:
        prepend.append("#include %s" % r)

    included = set()
    for line in lines:
        for p in prepend:
            if line.find(p) != -1:
                included.add(p);

    prepend = list(filter(lambda w: w not in included, [ p for p in prepend]))
    prepend.sort()

    for d in defines:
        prepend.append("#define %s" % d)

    if not using_namespace_std and requires:
        prepend.append("using namespace std;");




    return prepend




from __future__ import print_function

import re
import sys

from . import config

IGNORE_CHAR = '$'



def verbose(*args):
    if config.VERBOSE:
        debug(*args)

def debug(*args):
    print(' '.join(map(str, args)), file=sys.stderr)

def find_matching(line, open, close, start):
    j = start
    count = 1
    while j < len(line):
        if line[j] == close:
            count -= 1
            if count == 0:
                return line[start-1:j+1], j+1

        if line[j] == open:
            count += 1

        j += 1

    return line[start-1], start

# we split with split chars
# anything inside [], (), {} and <> are kept together
# anything inside "" or '' is kept together
# we pass a separator, like ' ' or ',' and we get back the inner pieces
def smart_split(line, split_chars=[], keep_splitters=False):
    split = []
    prev = []


    i = 0
    while i < len(line):
        c = line[i]
        i += 1

        if c in split_chars:
            if prev:
                split.append(''.join(prev))
                prev = []

            if keep_splitters:
                split.append(c)

        elif c == '"':
            p, i = find_matching(line, '"', c, i)
            prev.append(p)
        elif c == '<':
            p, i = find_matching(line, '<', '>', i)
            prev.append(p)
        elif c == '(':
            p, i = find_matching(line, '(', ')', i)
            prev.append(p)
        else:
            prev.append(c)

    if prev:
        split.append(''.join(prev))

    # debug("SPLIT", line, "INTO", split, "USING", split_chars)
    return split


# splits a 'b c' d into [a, 'b c', d]
def split_quotes(line):
#    return filter(lambda w: w, re.split(' (?=\[)',line))

    in_quotes = 0
    quote_char = None
    quotes = [ "'", '"' ]

    prev = []
    args = []
    for c in line:
        if c in quotes:
            if in_quotes:
                if c == quote_char and (not prev or prev[-1] != '\\'):
                    in_quotes -= 1
                    args.append('"%s"' % (''.join(prev)))
                    prev = []
                else:
                    prev.append(c)
            elif not in_quotes:
                quote_char = c
                in_quotes += 1
                if prev:
                    args.append(''.join(prev))
                    prev = []
        else:
            if c == ' ' and not in_quotes:
                if prev:
                    args.append(''.join(prev))
                prev = []
            else:
                prev.append(c)

    if prev:
        args.append(''.join(prev))

    return args

def get_indent(line):
    indent = 0
    for c in line:
        if c == ' ':
            indent += 1
        elif c == '\t':
            indent += 8
        elif c == IGNORE_CHAR:
            continue
        else:
            break

    return indent

def ptr_access(arg):
    return arg.find('->') != -1

def dot_access(arg):
    return arg.find('.') != -1

def array_access(arg):
    return arg.find('[') != -1

def visibility_line(line):
    return line.endswith('private:') or line.endswith('public:')

def hash_line(line):
    cline = line.strip()
    return cline and cline[0] == '#'


def is_def(line):
    return line.strip().find("def ") != -1

def is_struct(line):
    return (line.find(" struct ") != -1 or line.startswith("struct ")) and line.endswith('{')

def is_class(line):
    return line.strip().startswith("class ")

def line_is_include(line):
    return line.startswith("#include")

def line_is_using(line):
    return line.startswith("using ");

def line_is_template(line):
    line = line.strip().lstrip(IGNORE_CHAR).strip()
    return line.startswith("template<") or line.startswith("template <")

def case_statement(line):
    cline= line.strip()
    # default must have a : always
    return cline.startswith('case ')  or cline.startswith("default:")

def strip_outer_parens(s):
    if s[0] == '(' and s[-1] == ')':
        return s[1:-1]

    return s

def ignore_line(s, nl):
    sline = s.strip()
    if sline and sline[0] == IGNORE_CHAR:
        nl.append(s)
        return True

    return False

def extract_until_close(lines, i):
    count = 0
    while i < len(lines):
        line = lines[i]
        for c in line:
            if c == '{':
                count += 1
            if c == '}':
                count -= 1

        i += 1
        if count == 0:
            return i
if __name__ == "__main__":
    smart_split("int main(int argc, char **argv)", [' '])
    smart_split('"foo bar baz", int main', [','])
    smart_split('int main(vector <pair<int, int>> a, b', [','])
    smart_split("int i, j", [','])



# WARNING: THIS FILE IS DEPRECATED
# clanger is too slow (adds hundreds of ms per file). trying out custom code
# to recognize identifiers. the code is in id_recognizer.py

import clang.cindex
import os

from .util import *
IDX = None
def detect_clang_path():
    global IDX
    clang_paths = os.popen("locate libclang.so").read().split('\n')
    for path in clang_paths:
        clang.cindex.Config.set_library_file(path)
        try:
            IDX = clang.cindex.Index.create()
            debug("Set clang .so to", path)
            break
        except:
            continue

try:
    IDX = clang.cindex.Index.create()
except:
    debug("Couldn't create clang index, looking for .so files")
    detect_clang_path()

TokenKind = clang.cindex.TokenKind


# TODO: build a simple parser here that finds identifiers
# foo.obj.baz = '12'; should return [foo, '.', obj, '.', baz]
def parse_line(line):
    tu = IDX.parse('tmp.cpp', args=['-std=c++11'],
                    unsaved_files=[('tmp.cpp', line)],  options=0)
    for t in tu.get_tokens(extent=tu.cursor.extent):
        yield t


# we do some touchups to remove something like foo.bar.baz
# into only foo
def add_identifiers(line, scope):
    toks = parse_line(line)
    new = []

    new_toks = []
    toks = list(toks)
    while toks:
        t = toks.pop()
        new_toks.append(t)

        if toks:
            nt = toks[-1]
            if nt.spelling == '.' or nt.spelling == '->':
                toks.pop()
                new_toks.pop()


    for t in new_toks:
        if t.kind == TokenKind.IDENTIFIER:
            if t.spelling not in scope:
                new.append(t.spelling)
                scope[t.spelling] = 1

    return new

if __name__ == "__main__":
    scope = {}
    add_identifiers("vector<int> n", scope)
    add_identifiers("foo.bar. baz", scope)



from setuptools import setup

try:
    # python2
    execfile('okp/version.py')
except NameError as e:
    # python3
    eval('exec(open("./okp/version.py").read())')

setup(
    name='okp',
    version=__version__,
    author='okay',
    author_email='okayzed+okp@gmail.com',
    include_package_data=True,
    packages=['okp', 'okp.transforms'],
    scripts=['scripts/okp'],
    url='http://github.com/okayzed/okp',
    license='MIT',
    description='an compiler for .cpy files',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[ "future" ],
    )

